//Wir wollen in der lage sein commandozeilen parameter zu verarbeiten
#include <clang/Tooling/CommonOptionsParser.h>
//Wir sind eine frontend action
#include <clang/Frontend/FrontendActions.h>
//Wir machen Tooling
#include "clang/Tooling/Tooling.h"
//Wir benutzen die Visitor strategy
#include <clang/AST/RecursiveASTVisitor.h>
//Wir müssen mit dem compiler interagieren können.
#include <clang/Frontend/CompilerInstance.h>
#include <spdlog/spdlog.h>
#include <iostream>
#include "CTUApexAction.h"
#include "CTUApexConsumer.h"

#include "io/VersionTwoMCGReader.h"
#include "io/VersionThreeMCGReader.h"

#include "Util.h"
#include "ExtractionMetadata.h"


//Das sind convenience definitionen, dass wir nicht so viel tippen müssen
using namespace llvm;
using namespace clang;
using namespace clang::tooling;
using namespace clang;

//Die commandozeilen optionen die wir parsen wollen, gehören zu unserem program
static cl::OptionCategory CTUApex("Cross Translation Unit App Extraction");
//static cl::opt<std::string> callsites_file_opt("callsites-file", cl::desc("Config file with the callsites to outline"),
//                                               cl::value_desc("filename"), cl::Required, cl::cat(CTUApex));
static cl::opt<std::string> cg_file("cg-file", cl::desc("MetaCG file containing a whole programm call graph"),
                                    cl::value_desc("filename"), cl::Required, cl::cat(CTUApex));
enum class LogLevel {
    Trace, Debug, Info, Warning, Error, /* Critical,*/ Off
};

//static cl::opt<std::string> extractionStart("extract",
//                                            cl::desc("Qualified name of the function to start the extraction from"),
//                                            cl::value_desc("functionname"), cl::Optional, cl::cat(CTUApex));

static cl::list<std::string> extractionPoints("extract",
                                              cl::desc("Qualified names of the function to start the extraction from"),
                                              cl::value_desc("Comma separated list of function names"), cl::Optional,
                                              cl::cat(CTUApex));

static cl::opt<LogLevel> LoggingLevel("log-level", cl::desc("Select log level"),
                                      cl::values(clEnumValN(LogLevel::Trace, "trace", "Enable Trace and higher logs"),
                                                 clEnumValN(LogLevel::Debug, "debug", "Enable Debug and higher logs"),
                                                 clEnumValN(LogLevel::Info, "info", "Enable Info and higher logs"),
                                                 clEnumValN(LogLevel::Warning, "warning",
                                                            "Enable Warning and higher logs"),
                                                 clEnumValN(LogLevel::Error, "error", "Enable only Error logs"),
                                                 clEnumValN(LogLevel::Off, "off", "Disable all logging")),
                                      cl::init(LogLevel::Info), cl::cat(CTUApex));

static cl::opt<bool> OnlyWrapper("generate-only-wrapper", cl::desc("Output only the wrapper function"),
                                 cl::cat(CTUApex));
static cl::opt<bool> DumpAnalysisResult("dump-analysis-result",
                                        cl::desc("Dump the dependence analysis result to stdout"), cl::cat(CTUApex));

class CTUApexFrontendAction : clang::ASTFrontendAction {
public:
    CTUApexFrontendAction(metacg::Callgraph &cg) : cg(&cg) {}

    std::unique_ptr<clang::ASTConsumer> newASTConsumer() {
        return std::unique_ptr<clang::ASTConsumer>(new CTUApexConsumer(cg, includes, macros, pragmas));
    }


    std::unique_ptr<clang::ASTConsumer> CreateASTConsumer([[maybe_unused]] clang::CompilerInstance &compiler,
                                                          [[maybe_unused]] llvm::StringRef sr) {
        return std::unique_ptr<clang::ASTConsumer>(new CTUApexConsumer(cg, includes, macros, pragmas));
    }

private:
    std::vector<ExtractionBlock> includes;
    std::vector<ExtractionBlock> macros;
    std::vector<ExtractionBlock> pragmas;
    metacg::Callgraph *cg;
};


void markCalleesToExtract(const size_t nodeID, const metacg::Callgraph &cg) {
    for (const auto &call: cg.getCallees(nodeID)) {
        call->getOrCreateMD<ExtractionMetadata>()->setMarkedToExtract(true);
        markCalleesToExtract(call->getId(), cg);
    }
}

void markFunctionsAlongCallpaths(metacg::Callgraph &cg) {
    SPDLOG_INFO("Compute set of functions to be extracted");
    for (const auto &node: cg.getNodes()) {
        if (auto md = node.second->checkAndGet<ExtractionMetadata>();
                md.first && md.second->isMarkedToExtract()) {
            markCalleesToExtract(node.first, cg);
        }
    }
}

#include "clang/Basic/SourceManager.h"
#include "clang/Frontend/CompilerInstance.h"
#include "clang/Frontend/FrontendAction.h"
#include "clang/Lex/Preprocessor.h"
#include "clang/Lex/HeaderSearch.h"
#include "clang/Tooling/Tooling.h"

using namespace clang;
using namespace clang::tooling;

std::vector<std::string> clangIncludeBasedArgumentAdjustor(const CommandLineArguments & cla , StringRef sr){
    //call the compiler specified in cla[0] with -S and -v like so
    std::string command = "clang";
    //command.append(" -S -v");
    command.append("  -W,p -v -fsyntax-only");
    for(int i=1;i<cla.size();i++){
        command.append(" "+cla[i]);
    }
    command.append(" 2>&1");
    std::unique_ptr<FILE, decltype(&pclose)> pipe(popen(command.c_str(), "r"), pclose);
    //then read the output of the command like so:
    std::string txt;
    std::array<char, 128> buffer{};
    while (fgets(buffer.data(), buffer.size(), pipe.get()) != nullptr) {
        txt += buffer.data();
    }
    //search for "include search starts here" string and add all the paths with -I to the tool invocation
    txt=txt.substr(txt.rfind("search starts here:")+20,txt.rfind("End of search list.")-txt.rfind("search starts here:")-20);
    std::vector<std::string> ret;
    ret.assign(cla.begin(),cla.end());
    std::string::size_type pos;
    std::string::size_type prev = 0;
    while ((pos = txt.find('\n', prev)) != std::string::npos)
    {
        ret.push_back("-I"+txt.substr(prev, pos - prev));
        prev = pos + 2;
    }
    return ret;
}

std::vector<std::string> clangResourceDirBasedAdjuster(const CommandLineArguments & cla , StringRef){
    //call the compiler specified in cla[0] with -S and -v like so
    std::string command = "clang --print-resource-dir 2>1&";
    std::unique_ptr<FILE, decltype(&pclose)> pipe(popen(command.c_str(), "r"), pclose);
    //then read the output of the command like so:
    std::string txt;
    std::array<char, 128> buffer{};
    while (fgets(buffer.data(), buffer.size(), pipe.get()) != nullptr) {
        txt += buffer.data();
    }
    std::vector<std::string> ret;
    ret.assign(cla.begin(),cla.end());
    trim(txt);
    txt.pop_back();
    ret.push_back("-I"+txt+"/include");
    return ret;
}

int main(int argc, const char **argv) {

    auto ExpectedParser = CommonOptionsParser::create(argc, argv, CTUApex);

    if (!ExpectedParser) {
        // Fail gracefully for unsupported options.
        llvm::errs() << ExpectedParser.takeError();
        return 1;
    }

    switch (LoggingLevel) {
        case LogLevel::Trace:
            spdlog::set_level(spdlog::level::trace);
            break;
        case LogLevel::Debug:
            spdlog::set_level(spdlog::level::debug);
            break;
        case LogLevel::Info:
            spdlog::set_level(spdlog::level::info);
            break;
        case LogLevel::Warning:
            spdlog::set_level(spdlog::level::warn);
            break;
        case LogLevel::Error:
            spdlog::set_level(spdlog::level::err);
            break;
        case LogLevel::Off:
            spdlog::set_level(spdlog::level::off);
            break;
    }

    CommonOptionsParser &OptionsParser = ExpectedParser.get();


    ClangTool tool(OptionsParser.getCompilations(), OptionsParser.getSourcePathList());
    tool.clearArgumentsAdjusters();
    tool.appendArgumentsAdjuster(clangResourceDirBasedAdjuster);

    metacg::io::FileSource fs(cg_file.getValue());

    std::unique_ptr<metacg::io::MetaCGReader> mcgReader;

    if (fs.get().at("_MetaCG").at("version") == "3.0") {
        mcgReader = std::make_unique<metacg::io::VersionThreeMetaCGReader>(fs);
    } else if (fs.get().at("_MetaCG").at("version") == "2.0") {
        mcgReader = std::make_unique<metacg::io::VersionTwoMetaCGReader>(fs);
    } else {
        assert(false && "File corrupt or format not specified");
    }

    auto &mcgManager = metacg::graph::MCGManager::get();
    mcgManager.addToManagedGraphs("emptyGraph", std::make_unique<metacg::Callgraph>());
    nlohmann::json j = fs.get().at("_CG");
    std::cout<<j.dump(4);
    metacg::io::VersionTwoMetaCGReader::upgradeV2FormatToV3Format(j);
    std::cout<<j.dump(4);
    mcgReader->read();
    auto &cg = *mcgManager.getCallgraph();
    std::cout<<cg.size()<<"\n";

    //Add extraction point gotten via the CLI
    for(const auto& extractionFunc : extractionPoints){
        if (cg.hasNode(extractionFunc)) {
            auto *md = new ExtractionMetadata();
            md->setMarkedToExtract(true);
            cg.getNode(extractionFunc)->addMetaData(md);
        }else{
            SPDLOG_ERROR("The given function: "+ extractionFunc + " does not exist in the given callgraph");
        }
    }

    markFunctionsAlongCallpaths(cg);



    tool.run(clang::tooling::newFrontendActionFactory<CTUApexFrontendAction>(
            new CTUApexFrontendAction(*mcgManager.getCallgraph())).get());
    return 0;
}














